import * as socketIo from 'socket.io'
// import ReadLinePromised from "./util/ReadLinePromised";
const debug = require('debug')('ChatServer')

// we need to use a lock in order to avoid race conditions with the shared variables.
import AsyncLock = require('async-lock')
var lock = new AsyncLock()
import { MessageType, Message, MessageWithSeqNumber, ChatClientId, ChatConnectionId, ClientBrowserInfo } from '@stefancfuchs/chat-message-protocol-common'
import ChatServerConnection from './ChatServerConnection';

// export interface ChatConnection {
//   chatClientId?: ChatClientId
//   inputMessageChannel: InputMessageChannel
//   outputMessageChannel: OutputMessageChannel
//   inputControlChannel: InputControlChannel
//   outputControlChannel: OutputControlChannel
//   handshakeState: ChatHandshakeState
// }

interface ChatServerOptions {
  clientIdGenerator?: (args?: { widgetId: string, browserInfo: ClientBrowserInfo }) => Promise<ChatClientId> | ChatClientId,
  clientIdVerifier?: (args?: { clientId: ChatClientId, browserInfo: ClientBrowserInfo }) => Promise<boolean> | boolean,
  onClientConnected?: (args: { clientId: ChatClientId, connectionId: ChatConnectionId, chatConnection: ChatServerConnection, newClient: boolean, browserInfo: ClientBrowserInfo }) => void,
  onClientDisconnected?: (args: { clientId: ChatClientId, connectionId: number }) => void,
  onMessageReceived?: (args: { message: MessageWithSeqNumber, connectionId: ChatConnectionId, clientId: ChatClientId }) => void,
  // onClientRequestNewChat?: (args: { clientId: ChatClientId, connectionId: ChatConnectionId, chatConnection: ChatServerConnection }) => void
  // onClientNewBrowserInfo?: (clientId: ChatClientId, browserInfo: ClientBrowserInfo) => void
}

export default class ChatServer {
  public io: SocketIO.Server
  protected stopped: boolean = false
  protected port: number
  protected connections: Map<ChatConnectionId, ChatServerConnection> = new Map<ChatConnectionId, ChatServerConnection>()
  protected nextConnectionId: ChatConnectionId = 0
  protected nextClientId: number = 0
  // inputChannel: InputChannel
  // outputChannel: OutputChannel
  protected clientIdGenerator: ChatServerOptions['clientIdGenerator']
  protected clientIdVerifier: ChatServerOptions['clientIdVerifier']
  protected onClientConnected: ChatServerOptions['onClientConnected']
  protected onClientDisconnected: ChatServerOptions['onClientDisconnected']
  protected onMessageReceived: ChatServerOptions['onMessageReceived']
  // protected onClientRequestNewChat: ChatServerOptions['onClientRequestNewChat']

  constructor(port: number, options?: ChatServerOptions) {
    const {
      onClientConnected = undefined,
      onClientDisconnected = undefined,
      clientIdGenerator = () => "TEST_ID",
      clientIdVerifier = () => true,
      onMessageReceived = undefined,
      // onClientRequestNewChat = undefined
    } = options || {}
    this.port = port
    this.onClientConnected = onClientConnected
    this.onClientDisconnected = onClientDisconnected
    this.clientIdGenerator = clientIdGenerator
    this.clientIdVerifier = clientIdVerifier
    this.onMessageReceived = onMessageReceived
    // this.onClientRequestNewChat = onClientRequestNewChat

    this.io = socketIo(port)
    this.io.on('connect', async (socket: socketIo.Socket) => {
      debug('Connected client')
      if (this.stopped) {
        debug("The server is stopped")
        socket.disconnect()
        return
      }

      let connectionId: number
      await lock.acquire('key', () => {
        //FIXME theoretically an integer overflow can happen here
        connectionId = this.nextConnectionId++

        const chatConnection = new ChatServerConnection(socket, {
          clientIdGenerator,
          clientIdVerifier,
          connectionId,
          onMessageReceived: ({ message, connectionId, clientId }) => {
            debug(`message received:`)
            debug(message)
            this.onMessageReceived && this.onMessageReceived({ message, connectionId, clientId })
          },
          onHandshakeCompleted: ({ clientId, connectionId, newClient, browserInfo }) => {
            this.onClientConnected && this.onClientConnected({ clientId, connectionId, chatConnection, newClient, browserInfo })
          },
          /*
          onClientRequestNewChat: ({ clientId, connectionId }) => {
            this.onClientRequestNewChat && this.onClientRequestNewChat({ clientId, connectionId, chatConnection })
          },
          */
          // onControlMessageReceived: this.handleControlMessageReceived,
          onClientDisconnected: ({ clientId, connectionId }) => {
            this.onClientDisconnected && this.onClientDisconnected({ clientId, connectionId })
            this.connections.delete(connectionId)
          }
        })

        this.connections.set(connectionId, chatConnection)
      })
    })
  }

  // protected handleMessageReceived = (message: MessageWithSeqNumber, connectionId: ChatConnectionId, clientId: ChatClientId) => {
  //   const connection = this.connections.get(connectionId)
  //   if (!connection) return

  //   debug("Message received:")
  //   debug(message)

  //   if (connection.handshakeState === ChatHandshakeState.COMPLETED) {
  //     //process the message
  //     debug("Process message")
  //   }
  // }

  public getConnection = (index: number) => {
    return this.connections.get(index)
  }

  /**
   * Sends a message to client of the connection with the specified ID.
   * @returns The sequence number of the message. Null if the connectionId is invalid
   */
  public sendMessage = async (message: Message, connectionId: ChatConnectionId): Promise<number> => {
    const connection = this.connections.get(connectionId)
    if (!connection) {
      return null
    }
    return connection.sendMessage(message)
  }

  public sendBroadcastMessage = async (message: Message) => {
    this.connections.forEach(connection => {
      connection.sendMessage(message)
    })
  }

  /** Starts the server (server is started by default in constructor) */
  public start() {
    this.io = socketIo(this.port)
  }

  /** Stops the server */
  public stop() {
    this.stopped = true
    this.io.close()
    console.log(`[ChatServer] Disconnecting ${this.connections.size} clients...`)
    for (const [key, connection] of this.connections) {
      // console.log(`[ChatServer] Disconnecting client ${key}`)
      connection.socket.disconnect()
    }
    // return new Promise<void>(resolve => this.io.close(resolve))
  }

  // public consolePrompt = async () => {
  //   const readLine = new ReadLinePromised()
  //   while (true) {
  //     const str = await readLine.question(":")
  //     this.sendBroadcastMessage({ content: str, type: MessageType.TEXT })
  //   }
  // }


  /** Does a handshake with the client and returns the client ID.
   *  If the client is a new one, creates a new ID and returns it ot the client.
   *  If the client identifies itself as an existing one, calls the function checkClientId to check if the ID is valid.
   */
  // public async clientHandshake(connection: ChatConnection, checkClientId: (clientId: ChatClientId) => boolean): Promise<ChatClientId> {
  //   return String(this.nextClientId++)
  // }
}