import socketIo = require("socket.io");
const debug = require('debug')('ChatServerConnection')
import AsyncLock = require('async-lock')
import { ChatClientId, ChatConnectionId, MessageWithSeqNumber, OutputMessageChannel, ChatHandshakeState, Message, ControlMessageType, InputMessageChannel, InputControlChannel, OutputControlChannel, ClientBrowserInfo, ClientControlMessageContent } from "@stefancfuchs/chat-message-protocol-common"
var lock = new AsyncLock()



export interface ChatServerConnectionOptions {
  clientIdGenerator: (args?: { widgetId: string, browserInfo: ClientBrowserInfo }) => Promise<ChatClientId> | ChatClientId,
  clientIdVerifier: (args?: { clientId: ChatClientId, browserInfo: ClientBrowserInfo }) => Promise<boolean> | boolean,
  connectionId?: ChatConnectionId,
  onMessageReceived?: (args: { message: MessageWithSeqNumber, connectionId: ChatConnectionId, clientId: ChatClientId }) => void,
  onControlMessageReceived?: (args: { message: MessageWithSeqNumber, connectionId: ChatConnectionId, clientId: ChatClientId }) => void,
  maxBufferSize?: number,
  // onMessageAck?: (seqNumber: number) => void,
  onHandshakeCompleted?: (args: { clientId: ChatClientId, connectionId: ChatConnectionId, newClient: boolean, browserInfo: ClientBrowserInfo }) => void
  // onControlMessageReceived?: (controlMessage: ControlMessage, connectionId: ChatConnectionId) => void
  // onClientRequestNewChat?: (args: { clientId: ChatClientId, connectionId: ChatConnectionId, widgetId: string, browserInfo: ClientBrowserInfo }) => void
  onClientDisconnected?: (args: { clientId: ChatClientId, connectionId: ChatConnectionId }) => void
}

export default class ChatServerConnection {
  public socket: socketIo.Socket
  protected connectionId?: ChatConnectionId
  protected chatClientId?: ChatClientId
  protected inputMessageChannel: InputMessageChannel
  protected outputMessageChannel: OutputMessageChannel
  protected inputControlChannel: InputControlChannel
  protected outputControlChannel: OutputControlChannel
  protected handshakeState: ChatHandshakeState

  protected clientIdGenerator: ChatServerConnectionOptions['clientIdGenerator']
  protected clientIdVerifier: ChatServerConnectionOptions['clientIdVerifier']
  protected onHandshakeCompleted: ChatServerConnectionOptions['onHandshakeCompleted']
  // protected onClientRequestNewChat: ChatServerConnectionOptions['onClientRequestNewChat']
  protected onClientDisconnected: ChatServerConnectionOptions['onClientDisconnected']

  constructor(socket: socketIo.Socket, options?: ChatServerConnectionOptions) {
    const {
      clientIdGenerator = undefined,
      clientIdVerifier = undefined,
      connectionId = undefined,
      onMessageReceived = undefined,
      onControlMessageReceived = undefined,
      maxBufferSize = undefined,
      // onMessageAck = undefined,
      onHandshakeCompleted = undefined,
      // onClientRequestNewChat = undefined,
      onClientDisconnected = undefined,
    } = options || {}
    this.socket = socket
    this.handshakeState = ChatHandshakeState.START
    this.clientIdGenerator = clientIdGenerator
    this.clientIdVerifier = clientIdVerifier
    this.onHandshakeCompleted = onHandshakeCompleted
    this.connectionId = connectionId
    // this.onClientRequestNewChat = onClientRequestNewChat
    this.onClientDisconnected = onClientDisconnected

    //creates the message channels
    this.inputMessageChannel = new InputMessageChannel(socket, {
      connectionId,
      onMessageReceived: (message: MessageWithSeqNumber, connectionId: ChatConnectionId) => {
        if (this.handshakeState === ChatHandshakeState.COMPLETED) {
          onMessageReceived({ message, connectionId, clientId: this.chatClientId })
        }
      }
    })
    this.outputMessageChannel = new OutputMessageChannel(socket, { connectionId, maxBufferSize })
    this.outputMessageChannel.stopSendingMessages()

    //creates the control channels
    this.inputControlChannel = new InputControlChannel(socket, {
      connectionId,
      onMessageReceived: (message: MessageWithSeqNumber, connectionId: ChatConnectionId) => {
        this.handleControlMessageReceived(message)
        onControlMessageReceived && onControlMessageReceived({ message, connectionId, clientId: this.chatClientId })
      }
    })
    this.outputControlChannel = new OutputControlChannel(socket, { connectionId, maxBufferSize })

    socket.on('disconnect', () => {
      lock.acquire('key', () => {
        debug(`Client ${this.chatClientId} disconnected`)
        onClientDisconnected && onClientDisconnected({ clientId: this.chatClientId, connectionId: this.connectionId })
      })
    })
  }

  public sendMessage = async (message: Message): Promise<number> => {
    return this.outputMessageChannel.sendMessage(message)
  }

  protected handleControlMessageReceived = async (message: MessageWithSeqNumber) => {
    debug('handleControlMessage start')
    await lock.acquire('key', async () => {
      debug("Control message received:")
      debug(message)

      switch (message.type) {
        //The client requested a new ID
        case ControlMessageType.CLIENT_REQUEST_NEW_ID:
          if (this.handshakeState === ChatHandshakeState.START) {
            // const widgetId = message.content
            debug(`1 message.content: ${message.content}`)
            const contentObj: ClientControlMessageContent = JSON.parse(message.content)
            debug(`1 contentObj: ${contentObj}`)
            const widgetId = contentObj.widgetId
            const browserInfo = contentObj.browserInfo
            browserInfo.ip = this.socket.handshake.headers['x-forwarded-for'] || this.socket.handshake.address
            this.chatClientId = await this.clientIdGenerator({ widgetId, browserInfo })
            debug("Generated new client id: " + this.chatClientId)
            this.outputControlChannel.sendMessage({
              type: ControlMessageType.SERVER_ANSWER_NEW_CLIENT_ID,
              content: this.chatClientId
            })
            //TODO wait for ACK from client before completing handshake
            this.handshakeState = ChatHandshakeState.COMPLETED
            this.outputMessageChannel.startSendingMessages()
            this.onHandshakeCompleted && this.onHandshakeCompleted({ clientId: this.chatClientId, connectionId: this.connectionId, newClient: true, browserInfo })
          }
          break

        //The client has an existing ID
        case ControlMessageType.CLIENT_HAS_EXISTING_ID:
          if (this.handshakeState === ChatHandshakeState.START) {
            //Validate the client ID
            // const clientId = message.content
            debug(`1 message.content: ${message.content}`)
            const contentObj: ClientControlMessageContent = JSON.parse(message.content)
            debug(`1 contentObj: ${contentObj}`)
            const clientId = contentObj.clientId
            const browserInfo = contentObj.browserInfo
            browserInfo.ip = this.socket.handshake.headers['x-forwarded-for'] || this.socket.handshake.address
            if (await this.clientIdVerifier({ clientId, browserInfo })) {
              debug("client ID is valid")
              this.outputControlChannel.sendMessage({
                type: ControlMessageType.SERVER_ANSWER_CLIENT_ID_VALID,
                content: String(true) // tell the client the id is valid.
              })
              this.chatClientId = clientId
              this.handshakeState = ChatHandshakeState.COMPLETED
              this.outputMessageChannel.startSendingMessages()
              this.onHandshakeCompleted && this.onHandshakeCompleted({ clientId: this.chatClientId, connectionId: this.connectionId, newClient: false, browserInfo })
            } else {
              debug("client ID is invalid")
              this.outputControlChannel.sendMessage({
                type: ControlMessageType.SERVER_ANSWER_CLIENT_ID_VALID,
                content: String(false) // tell the client the id is invalid. It will then ask for a new one.
              })
            }
          }
          break

        // client is requesting a new chat
        /*
        case ControlMessageType.CLIENT_REQUEST_NEW_CHAT:
          const contentObj: ClientControlMessageContent = JSON.parse(message.content)
          const widgetId = contentObj.widgetId
          const browserInfo = contentObj.browserInfo
          this.onClientRequestNewChat && this.onClientRequestNewChat({ clientId: this.chatClientId, connectionId: this.connectionId, widgetId, browserInfo })
          break
        */
      }
    })
  }

}